;;; hamster-mode.el -- Modus fr Hamster-Filterdateien und Skripte
;; Copyright (C) 2002 Karl Pflsterer
;; Author: Karl Pflsterer
;; Created: Januar 2002
;; Version: 0.7
;; Keywords: mswin
;;; Commentary:

;; In dieser Datei sind einige, hoffentlich ntzliche Funktionen
;; gesammelt, zur Erleichterung der Erstellung und Bearbeitung von
;; Skriptdateiein und Scoredateien fr den Hamster.
;; Durch ein `(require 'hamster-mode) in der Startdatei oder
;; `(autoload 'hamster-mode "hamster-mode") und durch
;; `(push (cons '"\\.hsc" 'hamster-mode) auto-mode-alist)'
;; `(push (cons '"\\.hst" 'hamster-mode) auto-mode-alist)'
;; `(push (cons '"\\.hsm" 'hamster-mode) auto-mode-alist)'
;; wird `hamster-mode' beim ffnen von Dateien mit diesen Endungen
;; aktiviert. Einige Variablen stammen aus `hamster-functions.el'

;; Es sind einige Abkrzungen fr hufig verwendete Ausdrcke definiert.
;; Diese expandieren automatisch, wenn man den abbrev-minor-mode
;; aktiviert hat; alternativ wird die Expansion durch `C-x a e'
;; erzwungen.

;; C-c C-c ist an die Funktion `ham-mode-compile' gebunden. Diese
;; ruft den Hamster mit dem Namen des aktuellen Skriptes auf.

;; Die Einrckung erfolgt automatisch, nach Eingabe von `RETURN'
;; Der Wert der Einrckung wird durch die Variable
;; `ham-mode-default-indent' (default 2) festgelegt. `TAB' rckt eine
;; Zeile entsprechend dem Kontext ein.

;; Font-Locking ist fr Hamster-Skripte und Mail-, Newsfilterfiles
;; definiert.

;; `hamster-mode-hook' wird beim ffnen einer Skriptdatei aufgerufen.
;; Eine sinnvolle Definition dieses hooks in der Emacsstartdatei
;; knnte folgendermaen aussehen:

;; (add-hook 'hamster-mode-hook
;;	  (lambda()
;;	     (turn-on-font-lock)
;;	     (abbrev-mode 1)
;;	     (column-number-mode 1)
;;	     (line-number-mode 1)
;;	     (paren-toggle-open-paren-context 1)))

;;; Code:

(eval-when-compile
  (require 'font-lock))
(require 'hamster-functions)

;;; Variablen

(defconst hamster-exe "Hamster.exe "
  "*Name der Hamster-exe. Blanc am ende ist wichtig.")

(defconst ham-mode-scorefile-root-directory "Script/"
  "*Stammverzeichnis der HamsterSkripte.")

(defvar ham-mode-syntax-table nil
 "Syntax-Table im Hamster-Mode.")

(if ham-mode-syntax-table
    ()
  (setq ham-mode-syntax-table (make-syntax-table)))

(modify-syntax-entry ?$  "w   " ham-mode-syntax-table)
(modify-syntax-entry ?_  "w   " ham-mode-syntax-table)
(modify-syntax-entry ?#  "<   " ham-mode-syntax-table)
(modify-syntax-entry ?\\ "w   " ham-mode-syntax-table)
(modify-syntax-entry ?{  "<   " ham-mode-syntax-table)
(modify-syntax-entry ?}  ">   " ham-mode-syntax-table)
(modify-syntax-entry ?\n ">  " ham-mode-syntax-table)

(defvar ham-mode-map nil
  "Mode-spezifische Keymap.")

(if ham-mode-map
    ()
  (setq ham-mode-map (make-sparse-keymap))
  (define-key ham-mode-map "\t" 'ham-mode-hamster-tab)
  (define-key ham-mode-map "\r" 'ham-mode-electric-return)
  (define-key ham-mode-map "\M-\r" 'newline)
  (define-key ham-mode-map "\C-c\C-c" 'ham-mode-compile))

(defvar ham-mode-abbrev-table nil
  "Abbrev.-Table im Hamster-Mode.")

;; just some example abbreviations
(define-abbrev-table 'ham-mode-abbrev-table ())
(define-abbrev ham-mode-abbrev-table "hp" "HamPath")
(define-abbrev ham-mode-abbrev-table "hsp" "HamServerPath")
(define-abbrev ham-mode-abbrev-table "hgp" "HamGroupsPath")
(define-abbrev ham-mode-abbrev-table "hmp" "HamMailPath")
(define-abbrev ham-mode-abbrev-table "hnp" "HamNewsPath")

(defconst ham-mode-default-indent 2
  "*Default Einzug.")

(defvar hamster-mode-hook nil
  "Hook to run entering Hamster-Mode.")

(defconst ham-mode-keywords-1
  (concat  "while\\|ii?f\\|do\\|else\\|for\\|runscript\\|ini\\(read\\|write\\)"
	   "\\|end\\(?:ii?f\\|for\\|sub\\|while\\)"
	   "\\|HamWaitidle\\|execute\\|print\\|repeat\\|until\\|var\\|varset"
	   "\\|\\(Msg\\|Input\\|List\\|Popup\\|Text\\)Box"
	   "\\|inc\\|dec\\|icase\\|break"
	   "\\|Atclear\\|Atexecute\\|quit\\|return\\|Hammessage")
  "Keyworte fr Fontlock." )

(defconst ham-mode-keywords-2
  (concat "HamNews\\(Post||\Pull\\)\\|HamFetchMail\\|HamSendMail"
	  "\\|HamNewsJobs\\(Add\\|Clear\\|Delete\\|Feed\\|Post\\|" ;Vorsicht
	  "PostDef\\|Pull\\|PullDef\\|SetPrio\\|Start\\)" ;Klammer
	  "\\|HamNewsJobsGet\\(Counter\\|Param\\|Server\\|Type\\)"
	  "\\|HamGroup\\(Count\\|Name\\|NameByHandle\\|Index\\|Open\\|Close\\)"
	  "\\|FileExists\\|hamFlush"))

(defconst ham-mode-keywords-3
  (concat  "List\\(Add\\|Alloc\\|Box\\|Clear\\|Count\\|"
	   "Delete\\Dirs\\|Exists\\|Files\\|Free\\|Indexof\\|"
	   "Insert\\|Load\\|Raseentries\\|Save\\|Sort\\)"
	   "\\|List\\(Get\\|Set\\)\\(\\|Key\\|Tag\\|Text\\)"))

(defconst ham-mode-keywords-4
  (concat "RE_\\(Match\\|Extract\\|Parse\\<Split\\)"
	  "\\|len\\|pos\\copy\\|delete\\|trim\\|lowercase\\|"
	  "uppercase\\|replace\\|eval"))

;;; Font-Lock

(defvar ham-mode-font-lock-keywords nil)

(setq ham-mode-font-lock-keywords
;;;:***hsc ---------------------------------------------
      (list '("#.*\\|{.*}" 0 ham-mode-comment-face t)
	    '("\\(\\$[a-z]\\w*\\b\\)" 1 ham-mode-variable-face )
	    (list
	     "\\(#!\\(load\\|hs2\\|trace\\|debug\\|initialize\\)\\)[ \t]*\\([a-z_.]*\\)"
	     '(1 ham-mode-preprocessor-face t)
	     '(3 ham-mode-constant-face t))
	    (list "[ \t]*\\(sub\\) +\\(\\w+\\)"
		  '(1 ham-mode-keyword-face t)
		  '(2 ham-mode-constant-face))
	    (list "\\(label\\|goto\\)[ ]*([ ]*\\(\\w*\\)"
		  '(1 ham-mode-keyword-face)
		  '(2 ham-mode-constant-face))
	    (cons (concat
		   "\\<\\(?:"
		   "Ham\\(hsc\\|hsm\\|logs\\|server\\|"
		   "groups\\|mail\\|newsout\\|mailsout\\)?Path"
		   "\\|true\\|false"
		   "\\)\\>")
		   'ham-mode-constant-face)
	    (cons (concat
		   "\\<\\(?:"
		   ham-mode-keywords-1
		   "\\)\\>")
		  'ham-mode-keyword-face)
	    (cons (concat
		   "\\<\\(?:"
		   ham-mode-keywords-2
		   "\\)\\>")
		   'ham-mode-preprocessor-face)
	    (cons (concat
		   "\\<\\(?:"
		  ham-mode-keywords-3
		   "\\)\\>")
		  'ham-mode-list-function-face)
	    (cons (concat
		   "\\<\\(?:"
		   ham-mode-keywords-4
		   "\\)\\>")
		  'ham-mode-string-function-face)
	    (list "\\(Atadd[ ]*\\)(\\([ ]*[a-zA-Z,]*\\)"
		  '(1 ham-mode-keyword-face)
		  '(2 ham-mode-constant-face))
;; News -----------------------------------------------------
	    (list (concat "^=?\\(\\+\\|-\\)" ;1
			  "\\([0-9]*\\) +" ;2
			  "\\([~A-Za-z-]+\\) +"	;3
			  "\\({.+}\\|%[<>][0-9]+\\)") ;4
		  '(1 ham-mode-keyword-face)
		  '(2 ham-mode-constant-face)
		  '(3 ham-mode-keyword-face)
		  '(4 ham-mode-preprocessor-face t))
	    (list (concat "^=?\\(\\+\\|\\-\\)" ;1
			  "\\([0-9]*\\) +" ;2
			  "\\([~A-Za-z-]+\\) +") ;3
		  '(1 ham-mode-keyword-face)
		  '(2 ham-mode-constant-face)
		  '(3 ham-mode-keyword-face))
;; Mail ---------------------------------------------------
	    (list (concat "^=?\\(load()\\|ignore()\\|kill()\\)[ \t]+" ;1
			  "\\([-~A-Za-z]*\\)") ;2
		  '(1 ham-mode-keyword-face)
		  '(2 ham-mode-constant-face t))
	    (list (concat "^=?\\(load()\\|ignore()\\|kill()\\)[ \t]+" ;1
			  "\\([-~A-Za-z]*\\)" ;2
			  ": +"
			  "[-\\+]?\\(%[<>][0-9]+\\|{.+}\\)") ;3
		  '(1 ham-mode-keyword-face )
		  '(2 ham-mode-constant-face t)
		  '(3 ham-mode-preprocessor-face t))
	    '("Notify" 0 ham-mode-keyword-face t)))

(defvar ham-mode-start-block-regexp
  "^[ \t]*sub\\b\\|^[ \t]*if[ ]*(\\|^[ \t]*else\\b\\|^[ \t]*for[ \t]*(\\|^[ \t]*while[ \t]*("
"Regexp-Ausdruck, der Start eines Blockes kennzeichnet. ")


(defvar ham-mode-end-block-regexp
  "^[ \t]*\\bend\\(sub\\|if\\|for\\|while\\)\\b\\|^[ \t]*else\\b"
"Regexp-Ausdruck, der Ende eines Blockes kennzeichnet.")

;;; Faces

(defface ham-mode-preprocessor-face
  '((t (:foreground "OrangeRed" :bold nil))) "First group face")

(defface ham-mode-function-face
  '((t (:foreground "Red" :bold nil))) "First group face")

(defface ham-mode-list-function-face
  '((t (:foreground "indianred" :bold nil))) "First group face")

(defface ham-mode-comment-face
  '((t (:foreground "Blue" :bold nil))) "Second group face")

(defface ham-mode-keyword-face
  '((t (:foreground "Red" :bold nil))) "Second group face")

(defface ham-mode-constant-face
  '((t (:foreground "Steelblue" ))) "Second group face")

(defface ham-mode-variable-face
  '((t (:foreground "magenta4" ))) "Second group face")

(defface ham-mode-string-function-face
  '((t (:foreground "darkred" :bold nil))) "First group face")

;; ****************************************
;;; User Functions
;; ****************************************

(defun hamster-mode ()
  "Major Mode fuer Hamsterskripte und Filterdateien.
Spezielle Befehle: \\{ham-mode-map}."
  (interactive)
  (kill-all-local-variables)
  (use-local-map ham-mode-map)
  (make-local-variable 'parse-sexp-ignore-comments)
  (make-local-variable 'font-lock-defaults)
  (make-local-variable 'comment-start)
  (setq comment-start "#")
  (make-local-variable 'comment-start-skip)
  (make-local-variable 'comment-end)
  (setq comment-end "")
  (setq mode-name "Hamster"
	major-mode 'hamster-mode
	parse-sexp-ignore-comments t
	local-abbrev-table ham-mode-abbrev-table
	font-lock-defaults '(ham-mode-font-lock-keywords nil t))
  (set-syntax-table ham-mode-syntax-table)
  (run-hooks 'hamster-mode-hook))

(defun ham-mode-compile ()
  "Calls the hamster programm with the script in buffer as script. "
  (interactive)
  (let* ((datei (expand-file-name (buffer-file-name)))
	(datei (replace-in-string datei "/c/" "c:/")))
    (compile (concat ham-fun-root-directory
		     hamster-exe datei))))

;; Indentation

(defun ham-mode-hamster-tab ()
  "Indents the current line hopefully correctly. "
  (interactive)
  (save-excursion
    (beginning-of-line)
    (fixup-whitespace)
    (indent-to (ham-mode-calculate-indent))))

(defun ham-mode-electric-return ()
  "Combines `newline' and automagic indentation. "
  (interactive)
  (newline)
  (indent-to (ham-mode-calculate-indent)))

;; ****************************************
;;; Private Functions
;; ****************************************

(defun ham-mode-calculate-indent ()
  "Berechnet den Einzug der aktuellen Zeile."
  (save-excursion
    (let ((oldpos (point-at-eol))
	  (hmdi ham-mode-default-indent)
	  (indent 0)
	  (curpos (point)))
      (beginning-of-line 0)
      ;; Fortsetzungszeile ?
      (if (re-search-forward "_[\t ]*$"  (point-at-eol) t)
	(cond
	  ;; icase auf der aktuellen Zeile ?
	  ((search-backward "icase" (point-at-bol) t)
		(re-search-forward "\\$\\w+[ \t]*," (point-at-eol) t)
		(setq indent (+ 1 (current-column))))
	  ;; Fortsetzung einer Fortsetzung ?
	  ((re-search-backward "_[\t ]*$" (point-at-bol 0) t 2)
		(beginning-of-line 2)
		(re-search-forward  "[ \t]*" (point-at-eol) t)
		(setq indent (current-column)))
	  (t
	    (re-search-backward "^[ \t]*\\w+[ \t]*\\((\\|=\\)" nil t)
	    (re-search-forward "(\\|=")
	    (setq indent (- (current-column) 1))))
	(end-of-line 2)
	(while (re-search-backward  ham-mode-start-block-regexp (point-min) t)
	  (setq indent (+ indent hmdi)))
	(while (re-search-forward  ham-mode-end-block-regexp oldpos t)
	  (setq indent (- indent hmdi)))
	(goto-char curpos)
	(beginning-of-line)
	(if (looking-at ham-mode-start-block-regexp)
	  (setq indent (- indent hmdi))
	  indent)))))

(provide 'hamster-mode)

